/*-------------------------- =+- C / C++ -+= -------------------------------
 File:    cs547_pj.cc
 
 Created: Fri Dec 1 00:59:57 2000

 Yaroslav Halchenko				                             CS Dept. UNM
 e-mail:  yoh@cs.unm.edu                                     ICQ#: 60653192
 DESCRIPTION (NOTES):
 Future Project on "Attention". It takes set of digit patters and
 using attention signal computes error in back-prop net to feed whole
 image (28x28) in 8+2(attention_coordinates) channel and then recover
 it with min error
--------------------------------------------------------------------------*/

//#define PLOT

#define LOGING

#ifdef LOGING
# define LOG_ERR  "log_err.log"
# define LOG_SERR "log_serr.log"
#endif

//#define DEBUG

#define PATT_FROM_MONICA		// Takes patters from Monica's file
// #define printw printf

#ifdef PATT_FROM_MONICA
# define IN_FILE "input.txt"
#endif

#define USEADOLC				// First I thought about creating
//possibility to compute gradient not only using adolc library

#ifdef USEADOLC
# include <adolc.h>
#endif

#include "../dbase/dbase.h"		// to read digit-database
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <curses.h>				// to make possible interactive
//reaction on keys like 'g' - safe and show .gif of current pattern


#define PI (3.14)

#define FOR(i,n) for (int (i)=0; (i)<(n); (i)+=1)

#define CXY(i,j,d1,d2) ((i)*(d2)+j)	/* convert coordinates from 2D to
									   1D */
#define WXY(wp,wn,dp,dn) ((wn)*(dp)+wp)	/* in weights we keep in weird
										 form - i-s row keeps weights
										 to i-s unit of the next layer*/
#define POW2(x) ((x)*(x))

#define WIND(l1,l2) [(l1)*wbeg[l1]+l2] // computing of index in "w" matrix

#define IMS  (28)				/* size of the images side */
#define Ninp (IMS*IMS+3)		/* +1 is for 1 in bias, +2,3 center of
								   attention */
#define Nout Ninp				/* # of units in the output */
#define MAXNCYCLES (10000)		/* max number of cycles through
								   training data */
#define ATT_DEV (10)			/* deviatation of gaussian of attention */
#define ATT_I   (10)			// were used for fixed attention signal
#define ATT_J   (10)

#define Npat    (200)			// how many patterns to consider

#define NL (3)					/* number of layers */
const int Nu[NL]={ Ninp, 10, Nout }; /* number of units on each level */

/* Types of descent algorithm */
int descent_mode = 0;			// 0=SMD, 1=SGO 3 - stupid descent

#define WSIZE (Ninp*10+Nout*10)	// Size of w matrix

#include "meg/write_ppm.h"
#include "meg/stochastic-gradient.h"
#include "meg/random.h"
#include "meg/smd.h"

#ifdef USEADOLC
typedef adouble mdouble;
#else
typedef double mdouble;
#endif

#ifdef USEADOLC
adoublev  w(WSIZE);				// vector of w's for adolc
#else
mdouble*  w;					/* weights going to ith level*/
#endif
double*   wt;					// double w's
double*   dw;					

int       wbeg[NL];				// indeces of the beginning of ws for
								// each level in w, because they are
								// stored in linear array

double*   patterns[Npat];		// patterns are images of digits +
								// attention signals coordinates.

mdouble*  xs[NL];				/* all xs in my net - values of the
								   output of neurons on all levels */
mdouble*  xinp;					// inputs 
mdouble*  xout;					// outputs

mdouble   errs[Nout];			// errors
mdouble   att [Ninp];			// attention coefficients - again 2D
								// in 1D. So CXY() must be used


/* Simple 2D gaussian */
inline mdouble gaussian(mdouble i, mdouble ic) {
  return ((1/(sqrt(2*PI)*ATT_DEV)) * exp(-POW2((i-ic)/ATT_DEV)));
}

inline mdouble sigmoid(mdouble & x) {
  return (1 / (1 + exp(-x)));
}


/* Makes an array of coefficients to give weights to errors
   depending on the coordinates of the attention signal */
void makeAttention (mdouble ic, mdouble jc) {
  mdouble coef = 1/(gaussian(ic,ic) * gaussian(jc,jc)); 
  FOR(i, IMS)
	FOR(j, IMS)
	att[ CXY(i,j,IMS,IMS) ] = gaussian(i,ic) * gaussian(j,jc) * coef;
}

/* Forward pass. On output gives an error on that pass */
mdouble forwardPass( ){
  int starti;
  mdouble err=0, sum=0;
  
  FOR(l, NL-1) {
	FOR(xi, Nu[l+1]-3) {		// for all units besides bias (1) and
								// attention signals
	  
	  xs[l+1][xi] = 0;
	  starti = Nu[l] * xi;

	  sum=0;
	  FOR(xpi, Nu[l]) 
		sum += w WIND(l,starti++) * xs[l][xpi];
	  sum /= (Nu[l]);
	  xs[l+1][ xi ] = sigmoid(sum);
	  
	  xs[l+1][ Nu[ l+1 ]-2 ] = xs[0] [ Ninp - 2 ];
	  xs[l+1][ Nu[ l+1 ]-1 ] = xs[0] [ Ninp - 1 ];

	  xs[l+1][ Nu[ l+1 ]-3 ] = 1;
	}
  }

  
  FOR(i,Nout-3) {
	errs[i] = att[i] * (xout[i] - xinp[i]);	// take difference with
								// target which is an input
	err += sqrt(POW2(errs[i]));
  }
  
  err /= Nout;
  
  return (err);
}

/* I had a strange problem with fscanf() which couldn't read %f
   properly - so I had to write my own */
double readDouble(FILE *fi){
  char ch [100];
  int ind = 0;
  while(1) {
	ch[ind] = getc(fi);
	if ( (ch[ind] == EOF)) break;
	if ((ch[ind] == ' ')||(ch[ind] == '\t')||(ch[ind] == '\n') ) {
	  if (ind) break;
	}
	else
	  ind++;
  }
  ch[ind]=0;
  return(atof(ch));
}

/* Convert standart input - image from database to linear array and
   scale values to 1*/
void convertInput (timage & tim, double ares[], double ic, double jc ){
  FOR(i,IMS)
	FOR(j,IMS)
	ares[CXY(i,j,IMS,IMS)] = tim.A[i][j]/255.0;
  ares[Ninp-3] = 1;				/* bias */
  ares[Ninp-2] = ic/255.0;			/* attention coords */
  ares[Ninp-1] = jc/255,0;
}

/* Init all arrays etc */
void init_data(tdbase * db){
  FILE * fin;
  
  wbeg[0] = 0;
  FOR(i,NL) {
	//	starting indeces of ws in linear vector keeping ws
	wbeg[i+1] = wbeg[i] + Nu[i]*Nu[i+1];
  }
  if (wbeg[NL] != WSIZE) {
	fprintf(stderr,"Size of W (%d) is different from what supposed to be %d\n",
			wbeg[NL], WSIZE);
  }
  
  wt = new double [wbeg[NL]];
  dw = new double [wbeg[NL]];
  
  randomize_gauss((double *)wt, wbeg[NL], 0, 0.02);
  
#ifndef USEADOLC
  w = wt;
#else
#endif

#ifdef USEADOLC
  FOR(i,NL) xs[i] = new mdouble[Nu[i]];
#else
  FOR(i,NL-1) {					// we don't allocate for the first
								// level - it will be taken all the
								// time just by referencing pattern[npat]
	xs[i+1] = new mdouble[Nu[i+1]];
	//(mdouble *) calloc(Nu[i+1], sizeof(mdouble));
  }
#endif
  xout = xs[NL-1];				// point xout to last level

#ifdef PATT_FROM_MONICA
  if (!(fin = fopen(IN_FILE,"r"))) {
	printw("Can't read from file %c \n",IN_FILE);
	return;
  }
#endif
  
  FOR(i,Npat) {
	patterns[i] = new double [Ninp];
#ifndef PATT_FROM_MONICA
	convertInput(db->data[i % db->N], patterns[i], ATT_I, ATT_J);
#else
	{
	  double * ares = patterns[i];
	  double xx;
	  FOR(ii,IMS)
		FOR(j,IMS)
		ares[CXY(ii,j,IMS,IMS)] = readDouble(fin);
	  
	  ares[Ninp-3] = 1;				/* bias */
	  ares[Ninp-2] = readDouble(fin)*IMS/2+0.1;			/* attention coords */
	  ares[Ninp-1] = readDouble(fin)*IMS/2+0.1;
	  
	  FOR(ii,IMS)
		FOR(j,IMS)
		if (ares[CXY(ii,j,IMS,IMS)] != (xx=readDouble(fin))) {
		  printw("Target is not the same as source %e:%e\n",ares[CXY(ii,j,IMS,IMS)], xx);
		}
	}
#endif
  }
#ifdef PATT_FROM_MONICA
  fclose(fin);
#endif
  
}

/* Free memory used by our program */
void free_data(){
  delete [] wt;//free(w[i]);
  delete [] dw;
  
  FOR(i,NL-1) delete [] xs[i+1];//  free(xs[i+1]);
  FOR(i,Npat) delete [] patterns[i];//free(patterns[i]);
}

/* write weight to a file to make possible then to restore execution
   of the program not from the first epoch */
void write_weights(char filename[100]){
  FILE * fout;
  if (!(fout=fopen(filename,"w"))) {
	fprintf(stderr,"Can't open %s for writing\n",filename);
	return;
  }
  
  FOR(i,WSIZE)	fprintf(fout,"%3.10f ",wt[i]);
  
  fclose(fout);
}

/* read weights */
void read_weights(char* filename){
  FILE * fin;
  if (!(fin=fopen(filename,"r"))) {
	printw("Can't open %s for reading\n",filename);
	return;
  }
  FOR(i,WSIZE) {
	int dd,no;
	wt[i]=readDouble(fin);
	if (feof(fin)){
	  fprintf(stderr,"Bad news - file is over or an error\n");
	  return;
	}
  }
  fclose(fin);
}

/* Write .gif file of currently processed pattern + output + attention
   signal */
void writeGif(int ind, char display) {
  	char buf[1024];
	char image_file[100] = "";
	FILE *out_file;
	sprintf(image_file,"imgs/patt%d.gif",ind);
	sprintf(buf, "convert  - %s", image_file);

	if ( !(out_file = popen(buf, "w")) ) {
	  fprintf(stderr, "error: unable to open image output pipe '%s'\n", buf);
	  return ;
	}
	
	header_image(out_file, IMS*3, IMS);
	FOR(i,IMS) {
	  for(int l=0; l<=2*(NL-1); l=l+NL-1)
		FOR(j,IMS) {
		double d = (l<=(NL-1)?
					value(xs[l][CXY(i,j,IMS, IMS)]):
					value(att[CXY(i,j,IMS,IMS)]));
		int c = (int) rint(d*255);
		if (c>255) c=255;
		if (c<0) c=0;
		color_image(out_file,c,0,0);
	  }
	}
	
	pclose(out_file);

	if (display) {
	  sprintf(buf, "xv  %s &", image_file);
	  system (buf);
	}
}


/* Actually optimization routine */
void back_prop() {

  volatile double eta = 0.01;	// for stupid descent mode
  short  tag = 0;
  
  double Hv[WSIZE];
  int slen = (descent_mode == 0 ? smd_statelen(WSIZE) : sgo_statelen(WSIZE) );
  double state[slen];
#ifdef LOGING
  FILE * ferr, *fserr;
  if (!(ferr = fopen (LOG_ERR,"w")) || !(fserr = fopen (LOG_SERR,"w"))) {
	printw("Can't open log files\n");return;
  }
#endif
  
#ifdef PLOT
  char *  comml1="plotonline --title \"Error\" -x \"epoch\" -y \"error\" --no-point "; 
  FILE *plotout1;				// For plotonline
  /* open a pipe for plot */
  plotout1 = popen(comml1,"w");
#endif

  
  switch (descent_mode)    {
  case 0:
	smd_init(state, WSIZE);
	break;
  case 1:
	sgo_init(state, WSIZE);
	break;
  }

  FOR(epoch, MAXNCYCLES) {
	double err=0, serr=0;

	if(epoch && !(epoch%1)) {
	  char buff[100];
	
	  sprintf(buff,"weights%d_%d_%d_%d/epoch%d.dat",
			  NL, Nu[0], Nu[1], Nu[2], epoch);
	  write_weights(buff);
	}
	
	FOR(i, Npat) {
	  
	  char ch=-1;
	  ch = getch();				// serving interactive requests to
								// show .gif
	  if (ch!=EOF) {
		switch (ch){
		case 'g': printw("\n %c pressed - Gif image saved\n",ch);
		  writeGif(i, 1);
		  break;
		}
	  }
		
#ifdef USEADOLC
	  FOR(xx,Nu[0])
		xs[0][xx] = patterns[i][xx]; // assign pattern to input
	  
	  xinp = xs[0];				// point xinp to first level

	  makeAttention(patterns[i][Ninp-2], patterns[i][Ninp-1]);
		
	  trace_on(tag);
	  w <<= wt;
#else
	  xinp = patterns[i];
#endif
		
	  forwardPass() >>= serr;
	  err+= serr;
	  
#ifdef LOGING
	  fprintf(fserr, "%8.8g\n", serr);
#endif
	  
	  if (! (i%10)) {
		printw("Iter # %d training set image # %d error %3.6f\r",
			   epoch, i, serr);
#ifdef DEBUG
		printw(stderr,"Iter %d on pattern %d Error %3.6f\n",epoch,i,serr);
#endif
		fflush(stdout);
	  }
		
	  trace_off(tag);
		
	  gradient(tag, wbeg[NL-1], wt, dw);

	  if (descent_mode == 3) {
		FOR(xx, WSIZE)
		  wt[xx] -= eta * dw[xx];
	  }
	  else
		if (descent_mode == 0) {
		  if (epoch == 0) {
			smd_precondition(dw, wt, state, WSIZE, 0);
		  }
		  hess_vec(tag, WSIZE, wt, smd_state_v(state, WSIZE), Hv);
		  smd_step(dw, Hv, wt, state, WSIZE);
		} else {
		  sgo_step(dw, wt, state, WSIZE);
		}
	}
	printw("++++++Iter %d  Error %3.6f                               \n",epoch,err/Npat);
#ifdef LOGING
	fprintf(ferr, "%8.8g\n", err/Npat);
#endif
#ifdef PLOT
    fprintf(plotout1,"%8.8g\n",err/Npat);
    fflush(plotout1);    
#endif	
	refresh();
  }
  
#ifdef LOGING
  fclose(ferr);fclose(fserr);
#endif 
#ifdef PLOT
  fclose (plotout1);
#endif

}

/* MAIN FUNCTION */
int main (int argc, char* args[]) {
  tdbase training;
  unsigned int Nt=0,		/* number of images read from database */
    k, hm;
  char digits [10] = {1,1,1,0,0,0,1,0,1,1}; // which digits to read
  double E=0;
  
  initscr();  cbreak();  noecho(); // initializa curses
  keypad(stdscr, TRUE);  nodelay(stdscr, TRUE);
 

  printw("--------------------------------------------------------------------------------\n");
  printw("Reading Training Set of Images\n");
  
#ifndef PATT_FROM_MONICA
  Nt=read_dbase("../dbase/train-images-idx3-ubyte","../dbase/train-labels-idx1-ubyte", &training, digits);
  if (Nt < Npat)
	fprintf(stderr, "Number of read images less than number of  patterns supposed to be- some will be used twice - \n");
#endif
  
  init_data(&training);
  if (Nt) free_dbase(training);	// we don't need that data any more -
								// we have training data
								//  free_dbase(testing);
  
  if (argc==2) {
	printw("Read weights from file %s.....\n",args[1]);
	read_weights(args[1]);
  }
  printw("Training...........\n");
  back_prop();
  free_data();
  
  return 0;
}
